###############################################################################
package XAO::DO::Data::IndexData;
use strict;
use XAO::Objects;
use base XAO::Objects->load(objname => 'FS::Hash');
###############################################################################
1;
__END__

=head1 NAME

XAO::DO::Data::IndexData - default dynamic data object for Data::IndexData

=head1 SYNOPSIS

None

=head1 DESCRIPTION

The Data::IndexData object is derived from XAO::FS::Hash and does not
add any new methods.

=head1 AUTHORS

Copyright (c) 2003 XAO Inc.

Andrew Maltsev <am@xao.com>
